package org.qriterium.io;

import com.megginson.sax.*;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;

import org.xml.sax.SAXException;
import java.util.logging.Logger;
import java.util.zip.GZIPOutputStream;

import org.qriterium.statement.*;
import org.qriterium.util.StringUtilities;
import org.qriterium.util.XMLDataElement;
import org.qriterium.util.XMLProperties;

/** 
 * @author Timo
 * @author Arto
 * @author Erno
 */
public class QSPrinter implements StatementPrinter {
	private String filePath;
	private String fileName;
	private String fileSuffix;
	
	public final String DEFAULT_FILE_PATH = "";
	public final String DEFAULT_FILE_NAME = "QS";
	private static final String DEFAULT_FILE_SUFFIX = ".xml";
	
	private long outputSize;
	public final long DEFAULT_OUTPUT_SIZE = 0;
	private static final long OUTPUTSIZE_MARGIN = 1024;

	private static final String SERIALNUMBER_DELIM = "_part";

	private long serialNumber;
	private boolean gzip;
	
	private FileOutputStream outStream;
	private GZIPOutputStream gzipOutStream;
	private OutputStreamWriter outStreamWriter;
	public final String OUTPUT_CHARSET = "UTF-8";
	private DataWriter dataWriter;
	
	private static Logger logger = Logger.getLogger("org.qriterium.io.QSPrinter");;

	private static final int ZEROFILL_LENGTH = 4;
	
	private boolean outputActive;

	public QSPrinter() 
	{
		this.filePath = DEFAULT_FILE_PATH;
		this.fileName = DEFAULT_FILE_NAME;
		this.fileSuffix = DEFAULT_FILE_SUFFIX;

		this.outputSize = DEFAULT_OUTPUT_SIZE;
		this.serialNumber = 0;
		this.gzip = false;
	
		this.outStream = null;
		this.gzipOutStream = null;
		this.dataWriter = null;
		
		this.outputActive = false;
	}
			
	/**
	 * Prints one QStatement
	 */
	public void print(QStatement statement) {
		try {
			if (!this.outputActive) {
					fileStart();
					this.outputActive = true;
			}
			
			if (this.outputSize != 0) {
				this.dataWriter.flush();
				if (this.outStream.getChannel().size() + QSPrinter.OUTPUTSIZE_MARGIN > 
					this.outputSize) {
						fileEnd();
						fileStart();
				}
			}
			
			printStatement(statement);
		} catch (Exception e) {
			logger.severe("Printing failed: " + e.getMessage());
		}
	}
	
	/**
	 * Closes open streams
	 */
	public void close() {
		try {
			if (this.outputActive) {			
				fileEnd();
			}
			this.outputActive = false;
		} catch (Exception e) {
			logger.severe("Closing file failed: " + e.getMessage());
		}
	}

	/**
	 * Prints header for each output file
	 * 
	 * @throws SAXException
	 */
	private void fileStart() throws SAXException, IOException {
		this.serialNumber++;
		/*
		 * Open streams
		 */

		File outFile = new File(getCurrentFileName());
		
		this.outStream = new FileOutputStream(outFile);
		if (this.gzip) {
			this.gzipOutStream = new GZIPOutputStream(this.outStream);
			this.outStreamWriter = new OutputStreamWriter(this.gzipOutStream,
														  OUTPUT_CHARSET);
		} else {
			this.outStreamWriter = new OutputStreamWriter(this.outStream, 
														  OUTPUT_CHARSET);
		}

		this.dataWriter = new DataWriter(outStreamWriter);
		this.dataWriter.setIndentStep(XMLProperties.INDENT);

		/*
		 * Print header for each output file
		 */
		this.dataWriter.setPrefix(XMLProperties.NAMESPACE, 
								  XMLProperties.NAMESPACE_PREFIX);
		this.dataWriter.startDocument();
		this.dataWriter.startElement(XMLProperties.NAMESPACE, 
									 XMLProperties.ROOT);
	}
	
	/**
	 * Closes open tags for output file
	 * 
	 * @throws SAXException
	 * @throws IOException
	 */
	private void fileEnd() throws SAXException, IOException {
		/*
		 * Print footer for each output file
		 */
		this.dataWriter.endElement(XMLProperties.NAMESPACE, 
								   XMLProperties.ROOT);
		this.dataWriter.endDocument();
		
		/*
		 * Close streams
		 */
		if (this.gzip)
			this.gzipOutStream.finish();
	}
	
	/**
     * Qriterium String representation of the QStatement (format=XML)
     * @throws SAXException
     * @author Arto 
     */
	private void printStatement(QStatement statement) throws SAXException {
		this.dataWriter.startElement(XMLProperties.NAMESPACE, 
						 			 XMLProperties.STATEMENT);
		
		this.dataWriter.startElement(XMLProperties.NAMESPACE, 
						 			 XMLProperties.HEADER);
		
		this.printElements(statement.getHeader().getXMLDataElements());

		this.dataWriter.endElement(XMLProperties.NAMESPACE, 
					   			   XMLProperties.HEADER);
		
		this.dataWriter.startElement(XMLProperties.NAMESPACE, 
						 			 statement.getStartTag());

		this.printElements(statement.getXMLDataElements());

		this.dataWriter.endElement(XMLProperties.NAMESPACE, 
							   	   statement.getEndTag());
		
		this.dataWriter.endElement(XMLProperties.NAMESPACE, 
								   XMLProperties.STATEMENT);
	}

	/**
	 * Prints all elements in XMLDataElement array
	 * 
	 * @param elements
	 * @throws SAXException
	 */
	private void printElements(XMLDataElement[] elements) throws SAXException {
		for (int i = 0; i < elements.length; i++) {
			XMLDataElement e = elements[i];
			this.dataWriter.dataElement(e.getNamespace(), 
									    e.getName(), 
									    e.getValue());
		}
	}

	/**
	 * get current fileName
	 */
	public String getCurrentFileName() {
		String currentFileName;
		
		if (this.outputSize == 0)
			currentFileName = this.filePath + this.fileName + this.fileSuffix;
		else
			currentFileName = this.filePath + this.fileName + 
			           QSPrinter.SERIALNUMBER_DELIM + 
			           StringUtilities.zeroFill(this.serialNumber, ZEROFILL_LENGTH) +
			           this.fileSuffix;

		return currentFileName;
	}
	
	/*
	 * get/set methods
	 */
	
	public String getFilePath() {
		return this.filePath;
	}

	public void setFilePath(String filePath) {
		if (filePath.endsWith(java.io.File.separator)) {
			this.filePath = filePath;
		} else {
			this.filePath = filePath + java.io.File.separator;
		}
	}

	public String getFileName() {
		return this.fileName;
	}

	public void setFileName(String fileName) {
		if (fileName.matches("(.*)[.]xml$")) {
			fileName = fileName.substring(0, fileName.length() - 4);
		}
		this.fileName = fileName;
	}

	public boolean isGzip() {
		return this.gzip;
	}

	public void setGzip(boolean gzip) {
		if (gzip)
			this.fileSuffix = this.fileSuffix + ".gz";
		
		this.gzip = gzip;
	}

	public long getOutputSize() {
		return this.outputSize;
	}

	public void setOutputSize(long outputSize) {
		this.outputSize = outputSize;
	}

}
