package org.qriterium.statement;

import java.util.Date;

import org.qriterium.util.XMLDataElement;
import org.qriterium.util.XMLProperties;
import org.qriterium.util.IdFactory;

/**
 * Superclass for all statement-objects
 * @author Mikko
 * @author Arto
 */
public abstract class QStatement {

	protected QHeader header;
	
	/**
	 * Creates a new QStatement with creation date set to current date
	 * 
	 * @param hash the source hash for this statement
	 * @param rawId the raw id for this statement
	 */
	public QStatement(String hash, String rawId) {
		this(hash, rawId, new Date());
	}
	
	/**
	 * Creates a new QStatement
	 * @param hash the source hash for this statement
	 * @param rawId the raw id for this statement
	 * @param date creation date for this statement 
	 */
	public QStatement(String hash, String rawId, Date date) {
		this.header = new QHeader(IdFactory.getUniqueId(), 
		                                        hash,
		                                        rawId, 
		                                        date);
	}
	
	/**
	 * Set the header of this statement
	 * 
	 * @param header the header that will be set as the header for this
	 * statement
	 */
	public void setHeader(QHeader header) {
		this.header = header;
	}
	
	/**
	 * Returns the header of this statement
	 * @return the header of this statement
	 */
	public QHeader getHeader() {
		return this.header;
	}
	
	/**
	 * Returns the Qriterium id of this statement	
	 * @return the Qriterium id of this statement
	 */
	public long getId() {
		if (this.header == null) {
			return 0;
		}
		return this.header.getId();
	} 

	/**
	 * Returns the source hash of this statement	
	 * @return the source hash of this statement
	 */
    public String getHash() {
    	if (this.header == null) {
    		return null;
    	}
        return this.header.getHash();
    } 

	/**
	 * Returns the raw id of this statement	
	 * @return the raw id of this statement
	 */
    public String getRawId() {
    	if (this.header == null) {
    		return null;
    	}
        return this.header.getRawId();
    } 

	/**
	 * Returns the datestamp of this statement	
	 * @return the datestamp of this statement
	 */
    public Date getAt() {
    	if (this.header == null) {
    		return null;
    	}
    	return this.header.getAt();
    } 
    
    /**
     * Defines whether a QStatement can be connected by a QConnection with another QStatement.
     * Overwrite in subclasses if necessary.
     * @return true if QStatement can be a part of QConnection, otherwise false.
     */
    public boolean isConnectable() {
    	return false;
    }

     /**
     * XML Namespace of this QStatement.
     * Overwrite in subclasses if necessary.
     * @return XML namespace.
     */
    public String getXMLNamespace() {
    	return XMLProperties.NAMESPACE;
    }
    
    public String toString() {
    	return "QStatement <id: " + this.getId() + " rawId: " + this.getRawId() + ">";
    }
 
    /**
     * Gets XML data content of the QStatement 
     * @return XML data content of the QStatement.
     */
    public abstract XMLDataElement[] getXMLDataElements();

    /**
     * Returns the start tag of the QStatement, excluding namespace and tag-chars
     * Examples:
     * - "document"
     * - "content"
     * - "connection type="aware_of""
     * - "actor subtype="person""
     * @return start tag of the QStatement.
     * @author Arto
     */
	public abstract String getStartTag();

	/**
	 * Returns the end tag of the QStatement, excluding namespace and tag-chars
	 * @return end tag of the QStatement. For example "document"
	 */
	public abstract String getEndTag();
}
