package org.qriterium.transformer;

import java.io.*;
import java.util.Scanner;
import java.util.ArrayList;
import java.util.logging.Logger;
import java.net.URL;
import java.net.MalformedURLException;
import javax.xml.xpath.*;

import org.qriterium.connection.AwareOfConnection;
import org.qriterium.connection.ConsistsOfConnection;
import org.qriterium.connection.CreatedByConnection;
import org.qriterium.connection.ManagedByConnection;
import org.qriterium.connection.PublishedByConnection;
import org.qriterium.io.DataLeecher;
import org.qriterium.io.NullOutputStream;
import org.qriterium.statement.Organization;
import org.qriterium.statement.Person;
import org.qriterium.statement.QContent;
import org.qriterium.statement.QDocument;
import org.qriterium.statement.QStatement;
import org.qriterium.statement.QStatements;
import org.qriterium.statement.RawConnection;
import org.qriterium.util.IdFactory;
import org.qriterium.util.StatusPrinter;
import org.w3c.dom.*;
import org.xml.sax.*;

/**
 * This class is used for transforming Citeseer OAI-records to Qriterium
 * Statement-objects.
 * 
 * @author Mikko
 *
 */
public class CiteSeerTransformer extends Transformer {
	private static Logger logger = 
		Logger.getLogger("org.qriterium.transformer.CiteSeerTransformer");
	
	private Scanner scanner;
	
	/*
	 * the ArrayList where the statements parsed from a record will be
	 * placed before handing them over to the QStatements database
	 */
	private ArrayList<QStatement> statements;
	private ArrayList<RawConnection> rconnections;
	
	// the hash of this transformer
	private String transformer_id = "0";
	
	// the CiteSeer id for the record being parsed
	private String rawId;
	
	private int lineNumber;
	private int recordNumber;
	private int recordLineNumber;
	
	private long records_parsed;
	private long records_skipped;
	private long statements_created;

	/**
	 * Constructs a new CiteseerTransformer that produces Qriterium
	 * Statement-objects from the specified DataLeecher-interface
	 * 
	 * @param leecher A stream source implementing the Dataleecher-interface
	 * @param qStatements QStatements where to put the parsed statements into
	 */
	public CiteSeerTransformer() {		
		this.statements = new ArrayList<QStatement>();
		this.rconnections = new ArrayList<RawConnection>();
		
		this.transformer_id = IdFactory.getSerialID();
	}
	
	/**
	 * Starts creating Qriterium Statement-objects from the input stream
	 * and passes them to the QStatements-database 
	 *
	 */
	public void transform(DataLeecher dataLeecher, QStatements qStatements) {
		long currentTime = 0;
		long previousTime = 0;
		
		this.lineNumber = 1;
		this.recordNumber = 0;
		
		this.records_parsed = 0;
		this.records_skipped = 0;
		this.statements_created = 0;
		
		PrintStream stderr;
		PrintStream nullerr = new PrintStream(new NullOutputStream());
		
		this.scanner = new Scanner(dataLeecher.getInputStream(), "UTF8");
		
		StringReader sr = nextRecord();
		
		while (sr != null) {
			XPathFactory f = XPathFactory.newInstance();
			XPath xp = f.newXPath();
			InputSource is = new InputSource(sr);
			NodeList nodes = null;
			
			/*
			 * direct the stderr to null to suppress XPath
			 * error messages
			 */
			stderr = System.err;
			System.setErr(nullerr);
			
			// create a NodeList from the CiteSeer-record
			try {
				nodes = (NodeList)xp.evaluate("/record", is, 
										XPathConstants.NODESET);
			} catch (XPathExpressionException e){
				logger.warning("Skipped a non-well-formed record on line " + recordLineNumber + ".");
				this.records_skipped++;
				printStatus(false);
			}	
			
			// restore the stderr
			System.setErr(stderr);
			
			// only parse non-null NodeList
			if (nodes != null) {			
				/* 
				 * parse the NodeList into statements
				 * 
				 * if the record was parsed succesfully, 
				 * pass it to the QStatements-database
				 */
				if (parseRecord(nodes)) {
					this.records_parsed++;
					
					for (RawConnection rconnection : rconnections) {
						qStatements.addQStatement(rconnection);
						this.statements_created++;
					}
					for (QStatement statement : statements) {
						qStatements.addQStatement(statement);
						this.statements_created++;
					}
				} else {
					logger.warning("Skipped a malformed record on line " + recordLineNumber + ".");
					this.records_skipped++;
					printStatus(false);
				}
			}
			
			currentTime = System.nanoTime();
			if (currentTime - previousTime > 1000000000L) {
				printStatus(false);
				previousTime = currentTime;
			}
			
			sr = nextRecord();
		}
		
		// close the scanner when the inputstream has been read
		this.scanner.close();
		
		printStatus(true);
	}
	
	void printStatus(boolean permanent) {
		String msg = "Records processed : " + records_parsed + ", Records skipped : " + records_skipped;
		if (permanent)
			logger.info(msg);
		else
			StatusPrinter.printTemp(msg);	
	}
	
	/**
	 * Reads the next CiteSeer-record from the input stream and
	 * returns it as a StringReader object
	 * 
	 * @return a StringReader containing the next CiteSeer record
	 */
	private StringReader nextRecord() {
		StringBuffer record = new StringBuffer();
		String last = new String();
		
		this.recordNumber++;
		this.recordLineNumber = this.lineNumber;
		
		// read the inputstream until the next </record>-tag
		while (this.scanner.hasNextLine() && !last.equals("</record>")) {
			last = this.scanner.nextLine();
			record.append(last);
			this.lineNumber++;
		}
		
		// when we have reached the end of the record, return null
		if (record.length() != 0) {
			StringReader sr = new StringReader(record.toString());
			return sr;
		} else {
			return null;
		}
	}
	
	/**
	 * Parses the given CiteSeer record
	 * 
	 * @param nodes the record to be parsed
	 * @return true on success, false otherwise
	 */
	private boolean parseRecord(NodeList nodes) {
		// clear the previously parsed statements
		this.statements.clear();
		this.rconnections.clear();
		
		// start parsing if the record was read succesfully
		if (parseHeader(nodes) &&
				parseMetadata(nodes)) {
			return true;
		} else {
			return false;
		}
	}
	
	/**
	 * Parses the header from the given NodeList
	 * 
	 * @param nodes the nodelist to be parsed
	 * @return true on success, false otherwise
	 */
	private boolean parseHeader(NodeList nodes) {
		// read the header of the record
		Node idnode;
		
		try {
			idnode =
				nodes.item(0).getFirstChild().getChildNodes().item(0);
		} catch (NullPointerException e) {
			return false;
		}
			
		Node datenode = skipToNode(idnode, "datestamp");
		
		if (idnode != null &&
			idnode.getNodeName().equals("identifier") &&
			datenode != null) {
			rawId = idnode.getTextContent();
		} else {
			return false;
		}
		
		return true;
	}

	/**
	 * Parses the metadata section from the given NodeList
	 * 
	 * @param nodes the nodelist to be parsed
	 * @return true on success, false otherwise
	 */
	private boolean parseMetadata(NodeList nodes) {
		String title;
		URL openurl;
		String language;
		boolean publisherKnown;
		
		// the first child of the <oai_citeseer:oai_citeseer>-tag 
		Node mdnode;
		
		try {
			mdnode = nodes.item(0).getLastChild().
						getFirstChild().getFirstChild().
							getNextSibling();
		} catch (NullPointerException e) {
			return false;
		}
			
		// see if the metadata starts as we expect
		if (mdnode == null ||
			!mdnode.getNodeName().equals("dc:title")) {
			return false;
		}
			
		// title of the document		
		title = mdnode.getTextContent();
		
		// this will be used later to parse the authors
		Node authnode = mdnode; 
		
		// an Organization-statement from the publisher
		mdnode = skipToNode(mdnode, "dc:publisher");
		Organization pub;
		
		if (mdnode != null) {
			// only add the Organization if the publishers name
			// is not "unknown"
			pub = new Organization(mdnode.getTextContent(), this.transformer_id, rawId);
			
			if (!pub.getName().equals("unknown")) {
				statements.add(pub);
				publisherKnown = true;
			} else {
				publisherKnown = false;
			}
		} else {
			return false;
		}
			
		// read the openurl for the Document (<dc:identifier>)
		mdnode = skipToNode(mdnode, "dc:identifier");
			
		if (mdnode != null) {
			try {
				openurl = new URL(mdnode.getTextContent());
			} catch (MalformedURLException e) {
				return false;
			}
		} else {
			return false;
		}
			
		// read the language of the document
		mdnode = skipToNode(mdnode, "dc:language");
			
		if (mdnode != null) {
			language = mdnode.getTextContent();
		} else {
			return false;
		}
			
		// we can now construct the Document-statement
		//TODO filename != openUrl.toString() (??)
		QDocument qd = 
			new QDocument(title, language, openurl.toString(), this.transformer_id, rawId);
		statements.add(qd);
			
		/*
		 * now we can parse the authors and their affiliations
		 * and create the CreatedByConnections
		 */ 
		if (!parseAuthors(authnode, qd)) {
			return false;
		}
		
		// the content-statement
		QContent qc = new QContent(transformer_id, rawId);
		statements.add(qc);
		
		// ...and its ConsistsOfConnection to the document
		ConsistsOfConnection coc = 
			new ConsistsOfConnection(qd.getId(), qc.getId(), this.transformer_id, rawId);
		statements.add(coc);
		
		//a publishedByConnection to the publisher if it was known
		if (publisherKnown) {
			PublishedByConnection pbc = 
				new PublishedByConnection(qd.getId(), pub.getId(), this.transformer_id, rawId);
			statements.add(pbc);
		}
		// parse all the references
		if (!parseReferences(mdnode, qc)) {
			return false;
		}
		
		return true;
	}
	
	/**
	 * Parse authors and affiliations beginning from the given Node
	 * 
	 * @param mdnode the Node to start from
	 * @return true on success, false otherwise
	 */
	private boolean parseAuthors(Node mdnode, QDocument qd) {
		// skip to the first author
		mdnode = skipToNode(mdnode, "oai_citeseer:author");
		
		// make sure there is at least one author for this document
		if (mdnode != null) {
			Node authnode = mdnode;
			
			Person author;
			CreatedByConnection cbc;
			Organization org;
			ManagedByConnection mbc;
			
			do {
				String authname;
				
				if (authnode.hasAttributes()) {
					authname =
						authnode.getAttributes().item(0).getTextContent();
				} else {
					return false;
				}
				
				// create a Person-statement from the author
				author = new Person(authname, this.transformer_id, rawId);
				statements.add(author);
				
				// createdByConnection to the document
				cbc = new CreatedByConnection(qd.getId(), author.getId(),
											  this.transformer_id, rawId);
				statements.add(cbc);
						
				// read the possible affiliation tag
				Node affnode = 
					authnode.getLastChild();
				
				if (affnode != null) {
					affnode = affnode.getPreviousSibling();
					
					if (affnode != null &&
							affnode.getNodeName().equals("affiliation")) {
						String affiliation = affnode.getTextContent();
						
						// Organization-statement from the affiliation
						org = new Organization(affiliation, this.transformer_id, rawId);
						statements.add(org);
								
						/*
						 * create a managedByConnection between the
						 * actors 
						 */
						mbc = new ManagedByConnection(author.getId(), 
													  org.getId(),
													  this.transformer_id, rawId);
						statements.add(mbc);
					}
				}							
					
				authnode = skipToNode(authnode, "oai_citeseer:author");
			} while (authnode != null);
		
			return true;
		} else {
			return false;
		}
	}
	
	/**
	 * Parses all the CiteSeer references beginning from the given
	 * node and makes a rawconnection between them and the given
	 * QContent
	 * 
	 * @param mdnode the node to start from
	 * @param qc the QContent to connect the references to
	 * @return true on success, false otherwise
	 */
	private boolean parseReferences(Node mdnode, QContent qc) {
		// read all the possible references
		mdnode = skipToNode(mdnode, "oai_citeseer:relation");
		
		RawConnection rc;
		
		if (mdnode != null) {
			do {
				Node tonode = mdnode.getFirstChild();
				
				if (tonode != null) {
					tonode = skipToNode(tonode, "oai_citeseer:uri");
					
					if (tonode == null) {
						return false;
					}
				} else {
					return false;
				}
				
				/*
				 * create an awareOfConnection (raw) between the QContent
				 * and the target rawId
				 */
				rc = new RawConnection(qc.getRawId(), 
										tonode.getTextContent(),
										AwareOfConnection.TYPE,
										this.transformer_id, rawId);
				rconnections.add(rc);
				
				mdnode = skipToNode(mdnode, "oai_citeseer:relation");
			} while (mdnode != null);
		}
		
		return true;
	}
	
	/**
	 * Skips to the next sibling node with the given name starting from
	 * the given node.
	 * 
	 * @param node the node to start from
	 * @param tag the name of the wanted tag
	 * @return the node with the given tag
	 */
	private Node skipToNode(Node node, String tag) {
		try {
			node = node.getNextSibling();
			
			while (!node.getNodeName().equals(tag)) {
				node = node.getNextSibling();				
			}
		
			return node;
		} catch (NullPointerException e) {
			return null;
		}
	}
	
	/**
	 * Returns the amount of records this transformer has parsed
	 *  
	 * @return the amount of records this transformer has parsed
	 */
	public long getRecordsParsed() {
		return this.records_parsed;
	}

	/**
	 * Returns the amount of records this transformer has skipped
	 *  
	 * @return the amount of records this transformer has skipped
	 */
	public long getRecordsSkipped() {
		return this.records_skipped;
	}

	/**
	 * Returns the amount of statements this transformer has created
	 *  
	 * @return the amount of statements this transformer has created
	 */
	public long getStatementsCreated() {
		return this.statements_created;
	}
	
	public long getValidItems() {
		return this.records_parsed;
	}
	public long getInvalidItems() {
		return this.records_skipped;
	}
}
