package org.qriterium.transformer;

import java.io.*;
import java.util.Scanner;
import java.util.ArrayList;
import java.util.logging.Logger;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.xpath.*;

import org.qriterium.connection.AwareOfConnection;
import org.qriterium.connection.ConsistsOfConnection;
import org.qriterium.connection.CreatedByConnection;
import org.qriterium.connection.ManagedByConnection;
import org.qriterium.connection.PublishedByConnection;
import org.qriterium.io.DataLeecher;
import org.qriterium.io.NullOutputStream;
import org.qriterium.statement.Organization;
import org.qriterium.statement.Person;
import org.qriterium.statement.QContent;
import org.qriterium.statement.QDocument;
import org.qriterium.statement.QStatement;
import org.qriterium.statement.QStatements;
import org.qriterium.statement.RawConnection;
import org.qriterium.util.IdFactory;
import org.w3c.dom.*;
import org.xml.sax.*;


/**
 * This class is used for transforming DBLP records to Qriterium
 * Statement-objects.
 * 
 * @author Ilmari
 * @author Erno
 *
 */
public class DBLPTransformer extends Transformer {
	private static Logger logger = Logger.getLogger("org.qriterium.transformer.DBLPTransformer");
	
	// the hash of this transformer
	private String transformerId = "0";
	
	private ArrayList<QStatement> statements;
	private ArrayList<RawConnection> rconnections;
	
	private Scanner scanner;
	
	private int lineNumber;
	private int recordNumber;
	private int recordLineNumber;
	
	private long records_parsed;
	private long records_skipped;
	private int statementsCreated = 0;

	/**
	 * DBLP record starting tags.
	 * 
	 * http://dblp.uni-trier.de/xml/ 
	 */
	private final String[] STARTING_TAGS = {
		"article", "inproceedings", "proceedings", 
		"book", "incollection",
		"phdthesis", "mastersthesis", "www"
	};

	private final String[] AUTHOR_TAGS = {"author", "editor"};
	// private final String[] CONTACTED_AT_TAGS = {"address"};

	private final String[] MANAGER_TAGS = {"school"};
	private final String[] PUBLISHER_TAGS = {"school", "publisher", "journal"};
	// private final String[] PUBLISHING_DATE_TAGS = {"year", "month"}; 
	// private final String[] PUBLISHED_IN_TAGS = {
	//		"booktitle", "series", "journal", "crossref", 
	//		"volume", "number", "chapter", "pages" };

	private final String[] TITLE_TAGS = {"title"};
	private final String[] URI_TAGS = {"ee"}; // electronic edition
	// private final String[] IDENTIFIED_AS_TAGS = {"isbn", "url"}; // url is dblp metadata url
	private final String[] REFERENCE_TAGS = {"cite"};

	private QDocument qd;
	private QContent qc;
	private String rawId;
	
	/* document metadata tags (what are these anyhow?)
	"cdrom", 
	"note", 
	*/

	/**
	 * Constructs a new CiteseerTransformer that produces Qriterium
	 * Statement-objects from the specified DataLeecher-interface
	 * 
	 * @param leecher A stream source implementing the Dataleecher-interface
	 * @param qStatements QStatements where to put the parsed statements into
	 */
	public DBLPTransformer() {
		this.records_parsed = 0;
		this.records_skipped = 0;
		this.statementsCreated = 0;
		this.lineNumber = 1;
		this.recordNumber = 0;
		
		this.statements = new ArrayList<QStatement>();
		this.rconnections = new ArrayList<RawConnection>();
		
		this.transformerId = IdFactory.getSerialID();
	}
	
	/**
	 * Returns statement creation counter.
	 * 
	 * @return number of statements created
	 */
	public int getStatementsCreated()
	{
		return this.statementsCreated;
	}
	
	
	
	/**
	 * Starts creating Qriterium Statement-objects from the input stream
	 * and passes them to the QStatements-database 
	 *
	 */
	public void transform(DataLeecher dataLeecher, QStatements qStatements) {
		PrintStream stderr;
		PrintStream nullerr = new PrintStream(new NullOutputStream());
		
		this.scanner = new Scanner(dataLeecher.getInputStream(), "ISO-8859-1");

		StringReader sr = nextRecord();
		
		while (sr != null) {
			InputSource is = new InputSource(sr);
			Document nodes = null;
			
			/*
			 * direct the stderr to null to suppress XPath
			 * error messages
			 */
			stderr = System.err;
			System.setErr(nullerr);
			
			// create a NodeList from the DBLP record
			try {
				nodes = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(is);
			} catch (Exception e){
				this.records_skipped++;
				logger.warning("Skipped a non-well-formed record on line " + recordLineNumber + ".");
				logger.warning(e.getLocalizedMessage());
			}	
			
			// restore the stderr
			System.setErr(stderr);
			
			// only parse non-null NodeList
			if (nodes != null) {			
				/* 
				 * parse the NodeList into statements
				 * and pass it to the QStatements-database
				 */
				parseRecord(nodes); 
				this.records_parsed++;
				for (RawConnection rconnection : this.rconnections) {
					qStatements.addQStatement(rconnection);
				}
				for (QStatement statement : this.statements) {
					qStatements.addQStatement(statement);
				}
				this.statementsCreated += this.rconnections.size();
				this.statementsCreated += this.statements.size();
			}
			
			sr = nextRecord();
		}
		
		// close the scanner when the inputstream has been read
		scanner.close();
	}
	
	/**
	 * Reads the next DBLP record from the input stream and
	 * returns it as a StringReader object
	 * 
	 * @return a StringReader containing the next DBLP record
	 */
	private StringReader nextRecord() {
		StringBuffer record = new StringBuffer();
		String last = new String();
		
		this.recordNumber++;
		this.recordLineNumber = this.lineNumber;
		
		// read the inputstream until the next </STARTING_TAG>-tag
		while (scanner.hasNextLine() && !isTerminator(last)) {
			last = scanner.nextLine();
			if(!isHeader(last))
			{
				record.append(last);
			}
			this.lineNumber++;
		}
		
		// when we have reached the end of the record, return null
		if (record.length() != 0) {
			StringReader sr = new StringReader(record.toString());
			return sr;
		} else {
			return null;
		}
	}
	
	/**
	 * Detects record-ending tags.
	 * 
	 * @param line
	 * @return Whether the line is a record-ending tag or not.
	 */
	private boolean isTerminator(String line)
	{
		for(String tag: this.STARTING_TAGS)
			if(line.equals("</" + tag + ">"))
				return true;
		return false;
	}
	
	/**
	 * For detecting and ignoring the document root.
	 * 
	 * @param line
	 * @return Whether the line can be ignored or not.
	 */
	private boolean isHeader(String line)
	{
		return line.matches("<([!?]|/?dblp).*");
	}
	
	/**
	 * Parses the given CiteSeer record
	 * 
	 * @param nodes the record to be parsed
	 * @return true on success, false otherwise
	 */
	private boolean parseRecord(Document nodes) {
		// clear the previously parsed statements
		this.statements.clear();
		this.rconnections.clear();
		
		// start parsing if the record was read successfully
		return parse(nodes);
	}
	
	/**
	 * Parses the DBLP record from the given NodeList
	 * 
	 * @param nodes the nodelist to be parsed
	 * @return true on success, false otherwise
	 */
	private boolean parse(Document nodes) {
		ArrayList<String> 	authorNames = null, 
							managerNames = null,
							publisherNames = null,
							documentNames = null,
							urls = null,
							referenceRawIds = null;
		String  documentName = null,
				language = null,
				url = null;
		this.rawId = nodes.getDocumentElement().getAttribute("key");
		authorNames = extractTags(AUTHOR_TAGS, nodes);
		managerNames = extractTags(MANAGER_TAGS, nodes);
		publisherNames = extractTags(PUBLISHER_TAGS, nodes);
		documentNames = extractTags(TITLE_TAGS, nodes);
		referenceRawIds = extractTags(REFERENCE_TAGS, nodes);
		while(referenceRawIds.remove("..."));
		if(documentNames.isEmpty())
		{
			logger.warning("No document title found in record starting from line " + this.lineNumber + ".");
		} else {
			documentName = documentNames.get(0);
		}
		urls = extractTags(URI_TAGS, nodes);
		if(!urls.isEmpty())
			url = urls.get(0);

		logger.info(
			"New record: " + nodes.getDocumentElement().getTagName() + "\n" +
			"RawId: " + this.rawId + "\n" +
			"Authors: " + authorNames + "\n" +
			"Managers: " + managerNames + "\n" +
			"Publishers: " + publisherNames + "\n" +
			"Titles: " + documentNames + "\n" +
			"References: " + referenceRawIds + "\n" +
			"URLs: " + urls);

		createDocument(documentName, language, url);
		addAuthors(authorNames, managerNames);
		addPublishers(publisherNames);
		addReferences(referenceRawIds);
		
		return true;
	}

	private void createDocument(String title, String language, String filename) {
		this.qd = new QDocument(title, language, filename, transformerId, rawId);
		this.qc = new QContent(transformerId, rawId);
		ConsistsOfConnection coc = 
			new ConsistsOfConnection(qd.getId(), qc.getId(), transformerId, rawId);
		
		this.statements.add(this.qd);
		this.statements.add(this.qc);
		this.statements.add(coc);
	}
	
	private void addReferences(ArrayList<String> referenceRawIds) {
		if (referenceRawIds != null) {
			for (String referenceRawId: referenceRawIds) {
				RawConnection rc = new RawConnection(
						qc.getRawId(), 
						referenceRawId,
						AwareOfConnection.TYPE,
						transformerId, 
						rawId
				);
				this.rconnections.add(rc);
			}
		}
	}
	
	private void addAuthors(ArrayList<String> authorNames, ArrayList<String> managerNames) {
		ArrayList<Organization> managers = null;
		if(managerNames != null)
		{
			managers = new ArrayList<Organization>();
			for(String managerName: managerNames)
			{
				Organization manager = new Organization(
						managerName, transformerId, rawId);
				this.statements.add(manager);
				managers.add(manager);
			}
		}
		for(String authorName: authorNames)
		{
			Person author = new Person(authorName, transformerId, rawId);
			CreatedByConnection cbc = new CreatedByConnection(
				qd.getId(), author.getId(), transformerId, rawId);
			
			this.statements.add(author);
			this.statements.add(cbc);
			
			if(managers != null)
			{
				for(Organization manager: managers)
				{
					ManagedByConnection mbc = new ManagedByConnection(
							author.getId(), manager.getId(), 
							transformerId, rawId);
					this.statements.add(mbc);
				}
			}
		}
	}
	
	private void addPublishers(ArrayList<String> publisherNames) {
		if(publisherNames != null)
		{
			for (String publisherName: publisherNames)
			{
				Organization publisher = new Organization(
						publisherName, transformerId, rawId);
				PublishedByConnection pbc = new PublishedByConnection(
						qd.getId(), publisher.getId(), transformerId, rawId);
				
				this.statements.add(publisher);
				this.statements.add(pbc);
			}
		}
	}
	
	/**
	 * Extracts from nodes the text values of tags
	 * with the names specified in tags.
	 * 
	 * E.g. 
	 * [people]
	 * 		[man]Bob[/man]
	 * 		[woman]Jill[/woman]
	 * 		[dog]Woof[/dog]
	 * 	[/people]
	 * 
	 * extractTags {"man", "woman"} to that would return "Bob", "Jill".
	 * 
	 * @param tags
	 * @param nodes
	 * @return extracted tag values
	 * @throws Exception
	 */
	public ArrayList<String> extractTags(String[] tags, Document nodes)
	{
		ArrayList<String> values = new ArrayList<String>();
		XPath xp = XPathFactory.newInstance().newXPath();
		for(String tag: tags)
		{
			try {
				NodeList elems = (NodeList)xp.evaluate(
						"//"+tag+"/text()", nodes, XPathConstants.NODESET);
				for(int i=0; i<elems.getLength(); i++)
				{
					values.add(elems.item(i).getNodeValue());
				}
				xp.reset();
			} catch (Exception e) {
				logger.warning(e.getLocalizedMessage());
			}
		}
		return values;
	}

	public long getValidItems() {
		return this.records_parsed;
	}
	public long getInvalidItems() {
		return this.records_skipped;
	}
}
