package org.qriterium.transformer;

import java.util.logging.Logger;

import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.qriterium.io.DataLeecher;
import org.qriterium.statement.Organization;
import org.qriterium.statement.Person;
import org.qriterium.statement.QContent;
import org.qriterium.statement.QDocument;
import org.qriterium.statement.QStatements;
import org.qriterium.statement.RawConnection;
import org.qriterium.util.IdFactory;
import org.qriterium.util.SimpleNamespaceContext;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;

/**
 * This class is used for transforming existing Qriterium-XML to Qriterium
 * Statement-objects.
 * 
 * @author Mikko
 *
 */
public class QSTransformer extends Transformer {

	private static Logger logger = 
		Logger.getLogger("org.qriterium.transformer.CiteSeerTransformer");

	// the hash of this transformer
	private String transformer_id = "0";
	
	// how many statements has this transformer created
	private long statements_created;

	//private QStatements qStatements;
	
	/**
	 * Constructs a new QSTransformer that produces Qriterium
	 * Statement-objects from the specified DataLeecher-interface
	 * 
	 * @param leecher A stream source implementing the Dataleecher-interface
	 * @param qStatements QStatements where to put the parsed statements into
	 */
	public QSTransformer() {
		this.statements_created = 0;
		
		this.transformer_id = IdFactory.getSerialID();
	}

	/**
	 * Starts creating Qriterium Statement-objects from the input stream
	 * and passes them to the QStatements-database 
	 */
	public void transform(DataLeecher dataLeecher, QStatements qStatements) {
		this.qStatements = qStatements;
		
		XPathFactory f = XPathFactory.newInstance();
		XPath xp = f.newXPath();
		
		// set the correct namespace
		SimpleNamespaceContext nci = 
			new SimpleNamespaceContext("qs", "http://www.qriterium.org/");
			
		xp.setNamespaceContext(nci);
		
		InputSource in = new InputSource(dataLeecher.getInputStream());
		NodeList nodes = null;
		
		// create a NodeList from the statement-list
		try {
			nodes = (NodeList)xp.evaluate("/qs:statement_list/qs:statement", 
											in, XPathConstants.NODESET);
		} catch (XPathExpressionException e){
			logger.warning("The input is not well-formed!");
		}
		
		// if the statement list was parsed successfully
		if (nodes != null) {
			for (int i = 0; i < nodes.getLength(); i++) {
				parseNode(nodes.item(i));
			}
		}
		
	}
	
	// parses a single node
	private void parseNode(Node node) {
		String type = node.getChildNodes().item(3).getNodeName();
		NodeList bodyNodes = node.getChildNodes().item(3).getChildNodes();	
		String rawId = node.getChildNodes().item(1).getChildNodes().item(1).getTextContent();
		
		// parse nodes according to statement type
		if (type.equals("qs:document")) {
			parseDocument(rawId, bodyNodes);
		
		} else if (type.equals("qs:content")) {
			parseContent(rawId, bodyNodes);
		
		} else if (type.equals("qs:actor")) {
			// parse according to the type of the actor
			String attr = 
				node.getChildNodes().item(3).getAttributes().item(0).getNodeValue();
			
			if (attr.equals("person")) {
				parsePerson(rawId, bodyNodes);
		
			} else if (attr.equals("organization")) {
				parseOrganization(rawId, bodyNodes);
			
			} 
			
		} else if (type.equals("qs:connection")) {
			String connection_type = 
				node.getChildNodes().item(3).getAttributes().item(0).getNodeValue();
			
			parseConnection(connection_type, rawId, bodyNodes);
		}
	}
	
	// all connections must be added as RawConnections
	private void parseConnection(String connection_type, String rawId, NodeList bodyNodes) {
		RawConnection statement = new RawConnection(
				bodyNodes.item(1).getTextContent(),
				bodyNodes.item(3).getTextContent(),
				connection_type,
				this.transformer_id, rawId);
		this.qStatements.addQStatement(statement);
		this.qStatements.checkConnectionToPoint(statement);
		this.statements_created++;
	}

	private void parseOrganization(String rawId, NodeList bodyNodes) {
		Organization statement = new Organization(
				bodyNodes.item(1).getTextContent(),
				this.transformer_id, rawId);
		this.qStatements.addQStatement(statement);
		this.qStatements.addConnectionPoint(statement);
		this.statements_created++;
	}

	private void parsePerson(String rawId, NodeList bodyNodes) {
		Person statement = new Person(
				bodyNodes.item(1).getTextContent(),
				this.transformer_id, rawId);
		this.qStatements.addQStatement(statement);
		this.qStatements.addConnectionPoint(statement);
		this.statements_created++;
	}

	private void parseContent(String rawId, NodeList bodyNodes) {
		this.qStatements.addQStatement(new QContent(this.transformer_id, rawId));
		this.statements_created++;
		// Content is always a connection point
	}

	private void parseDocument(String rawId, NodeList bodyNodes) {
		QDocument statement = new QDocument(
				   bodyNodes.item(1).getTextContent(),
				   bodyNodes.item(3).getTextContent(),
				   bodyNodes.item(5).getTextContent(),
				   this.transformer_id, rawId);
		this.qStatements.addQStatement(statement);
		this.qStatements.addConnectionPoint(statement);
		this.statements_created++;
	}

	/**
	 * Returns the amount of statements this transformer has created
	 *  
	 * @return the amount of statements this transformer has created
	 */
	public long getStatementsCreated() {
		return this.statements_created;
	}
	
	public long getValidItems() {
		return this.statements_created;
	}
	public long getInvalidItems() {
		return 0;
	}
}
