package org.qriterium.ui;

// import java.io.BufferedOutputStream;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
// import java.io.FileNotFoundException;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.PrintWriter;
import java.util.Date;
import java.util.logging.Handler;
import java.util.logging.Logger;
import java.util.logging.Level;

import jargs.gnu.CmdLineParser;

import org.qriterium.util.IdFactory;
import org.qriterium.util.LogFormatter;
import org.qriterium.util.LogHandler;

import org.qriterium.connection.ManagedByConnection;
import org.qriterium.io.ConsolePrinter;
import org.qriterium.io.DataLeecher;
import org.qriterium.io.FileReader;
import org.qriterium.io.QSPrinter;
// import org.qriterium.io.StatementPrinter;
import org.qriterium.statement.Person;
import org.qriterium.statement.QSource;
import org.qriterium.statement.QStatements;
import org.qriterium.transformer.*;

import org.qriterium.settings.*;

/**
 * CLI user interface for Boa Transformer
 * 
 * @author Timo
 * @author Erno
 */

public class Boa {
	private static final int ARGUMENT_ERROR = 1;
	// private static final int NO_TRANSFORMERS_SPECIFIED_ERROR = 2;
	private static final int UNKNOWN_TRANSFORMER_SPECIFIED_ERROR = 3;
	private static final int NO_INPUT_FILES_ERROR = 4;
	// private static final int OUTPUT_FILE_EXISTS_ERROR = 5;
	private static final int TMP_DIRECTORY_NOT_WRITABLE = 6;
	private static final int ANOTHER_INSTANCE_RUNNING_ERROR = 7;
	
	private static Logger logger = Logger.getLogger("org.qriterium");

	private static String mainConfig;
	private static boolean verbose;
	private static boolean help;

	/**
	 * @param args
	 *            Command line arguments
	 */
	public static void main(String[] args) {
		/* check and create a single instance lock */
		lock();
		
		parseArguments(args);

		// -h command line option
		if (help) {
			printUsage();
			System.exit(0);
		}

		// -v command line option
		if (verbose)
			initializeLogger(Level.INFO);
		else
			initializeLogger(Level.SEVERE);
		
		Loader.loadSettings(mainConfig);
		
		String idStateFile = Loader.getDataDir() + File.separator + "IdFactory.state";
		
		if (Loader.getDataDir() == null) 
			System.exit(NO_INPUT_FILES_ERROR);
		
		IdFactory.retrieve(idStateFile);
		
		for (Source source : Loader.getSources()) {
			logger.info("Processing source    : " + source.getName());
			logger.info("    User             : " + source.getUser().getName());
			logger.info("    Input Directory  : " + source.getInputDir());
			logger.info("    Output Directory : " + source.getOutputDir());
			logger.info("    Output File      : " + source.getOutputFile());
			logger.info("    Data Directory   : " + source.getDataDir());
			logger.info("    Transformer      : " + source.getTransformer());
			logger.info("    Printer          : " + source.getPrinter());
			logger.info("    Cut size         : " + source.getCutSize());
			logger.info("    Gzip             : " + source.getGzip());

			SourceState sourceState = new SourceState();
		
			// restore sourceState object
			String sourceStateFile = source.getDataDir() + File.separator + source.getFilename() + ".state";
			try {
				FileInputStream fileStream = new FileInputStream(sourceStateFile);
				ObjectInputStream objectStream = new ObjectInputStream(fileStream);
				sourceState = (SourceState) objectStream.readObject();
				objectStream.close();
				fileStream.close();
			} catch (FileNotFoundException e) {
				logger.info("No previous state for this source.");
			} catch (Exception e) {
				logger.severe("Couldn't read previous state for this source.");
			}
			
			logger.info("  State:");
			logger.info("    NextFileID       : " + sourceState.getNextFileID());
			logger.info("    NextSerialID     : " + sourceState.getNextSerialID());
			
			QStatements statements = new QStatements();

			String indexStateFile = source.getDataDir() + File.separator + source.getFilename() + ".index.state";
			String connectionStateFile = source.getDataDir() + File.separator + source.getFilename() + ".connection.state";
			
			statements.getIndex().retrieve(indexStateFile);
			statements.getQConnectionHandler().retrieve(connectionStateFile);
			
			/* initialize correct transformer */
			Transformer transformer = null;
			
			if (source.getTransformer().equalsIgnoreCase("citeseer")) {
				transformer = new CiteSeerTransformer();
			}
			else if (source.getTransformer().equalsIgnoreCase("dblp")) {
				transformer = new DBLPTransformer();			
			}
			else if (source.getTransformer().equalsIgnoreCase("qs")) {
				transformer = new QSTransformer();			
			}
			else {
				logger.severe("Unknown transformer specified: " + 
						source.getTransformer());
				System.exit(UNKNOWN_TRANSFORMER_SPECIFIED_ERROR);
			}
			
			/* initialize correct printer */
			if (source.getPrinter().equalsIgnoreCase("qs")) {
				QSPrinter printer = new QSPrinter();
				printer.setFilePath(source.getOutputDir());
				
				String outputFilename;
				if (source.getOutputFile() != null) {
					outputFilename = source.getOutputFile(); 
				} else {
					outputFilename = "QS";
				}
				
				outputFilename = outputFilename + "." + sourceState.getNextFileID();
				printer.setFileName(outputFilename);
				logger.info("Writer output set to file : " + outputFilename);
				
				printer.setGzip(source.getGzip());
				printer.setOutputSize(source.getCutSize());
				statements.addPrinter(printer);			
			}
			else if (source.getPrinter().equalsIgnoreCase("console")) {
				ConsolePrinter printer = new ConsolePrinter();
				statements.addPrinter(printer);
			}
			
			logger.info("Processing files in directory : " + source.getInputDir());
			
			/* run transformer for files in inputDir */
			File inputDir = new File(source.getInputDir());
			String[] files = inputDir.list();
			
			// create user and source statements in the first file
			if (sourceState.getSourceID() == 0) {
				QSource qsSource = new QSource(source.getName(), null, IdFactory.getSerialID(), "");
				sourceState.setSourceID(qsSource.getId());
				Person qsPerson = new Person(source.getUser().getName(), IdFactory.getSerialID(), "");
				statements.addQStatement(qsSource);
				statements.addQStatement(qsPerson);
				statements.addQStatement(new ManagedByConnection(qsSource.getId(), qsPerson.getId(), IdFactory.getSerialID(), ""));
			}
			
			boolean newFiles = false;
			
			// process all files in input directory
			for (String filename : files) {
				if (sourceState.isCompleted(filename)) {
					logger.info("Skipping already processed file : " + filename);
					continue;
				}
				
				String fullname = source.getInputDir() + File.separator + filename;
				if (new File(fullname).isDirectory())
					continue;
				
				logger.info("Processing input file : " + fullname);
				try {
					DataLeecher leecher = new FileReader(fullname);
					transformer.transform(leecher, statements);
				} catch (Exception e) {
					logger.severe("Cannot read file : " + fullname);
				}
				
				FileState fileState = new FileState();
				fileState.setCreationID(IdFactory.getState());
				fileState.setDate(new Date());
				fileState.setValidItems(transformer.getValidItems());
				fileState.setInvalidItems(transformer.getInvalidItems());
				
				sourceState.setFileState(filename, fileState);
				
				newFiles = true;
			}
			
			statements.finishPrinters();
			
			// store state files
			statements.getIndex().store(indexStateFile);
			statements.getQConnectionHandler().store(connectionStateFile);

			if (newFiles)
				sourceState.setNextFileID();
			
			try {
				FileOutputStream fileStream = new FileOutputStream(sourceStateFile);
				ObjectOutputStream objectStream = new ObjectOutputStream(fileStream);
				objectStream.writeObject(sourceState);
				objectStream.close();
				fileStream.close();
			} catch (Exception e) {
				logger.severe("Error while writing state file.");
			}
			
			sourceState.updateLastRun();
			
			// Create summary file
			if (source.getOutputFile() != null && source.getOutputDir() != null) {
				try {
					PrintWriter out = new PrintWriter(
						new BufferedWriter(
							new FileWriter(source.getOutputDir() + File.separator + source.getOutputFile() + ".txt")
						)
					);
					out.println("Source      : " + source.getName());
					out.println("Transformer : " + source.getTransformer());
					out.println("Printer     : " + source.getPrinter());
					out.println("Gzip        : " + source.getGzip());
					out.println("Cut size    : " + source.getCutSize() + " bytes");
					out.println("Last run    : " + sourceState.getLastRun());
					out.println();
					out.println("Input files :");
					for (String filename : sourceState.getFileNames()) {
						out.println(filename);
						FileState fileState = sourceState.getFileState(filename);
						out.println("   File was processed : " + fileState.getDate());
						out.println("   Valid items        : " + fileState.getValidItems());
						out.println("   Invalid items      : " + fileState.getInvalidItems());
					}
					out.close();
				} catch (Exception e) {
					System.out.println(e);
				}
			}
		}
		
		IdFactory.store(idStateFile);
	}

	/**
	 * Handles parsing of command line arguments
	 * @param args command line arguments
	 */	
	private static void parseArguments(String[] args) {
		CmdLineParser cmdParser = new CmdLineParser();
		CmdLineParser.Option optVerbose = cmdParser.addBooleanOption('v', "verbose");
		CmdLineParser.Option optMainConfig = cmdParser.addStringOption('m', "mainconfig");
		CmdLineParser.Option optHelp = cmdParser.addBooleanOption('h', "help");

		try {
			cmdParser.parse(args);
		} catch (CmdLineParser.OptionException e) {
			System.err.println("Command line parser failed: " + e.getMessage());
			printUsage();
			System.exit(ARGUMENT_ERROR);
		}
		
		mainConfig = (String) cmdParser.getOptionValue(optMainConfig);
		verbose = (Boolean) cmdParser.getOptionValue(optVerbose, Boolean.FALSE);
		help = (Boolean) cmdParser.getOptionValue(optHelp, Boolean.FALSE);
	}
		
	/**
	 * Initializes the Logger object to use
	 * 
	 * @param level
	 *            the level of messages to display
	 */
	private static void initializeLogger(Level level) {
		Handler handler = new LogHandler();
		handler.setFormatter(new LogFormatter());
		logger.addHandler(handler);
		logger.setUseParentHandlers(false);
		logger.setLevel(level);
	}

	/**
	 * Function to check and create a lock file so that the program can be run
	 * only one instance at a time.
	 * 
	 * Note that the lock file is removed automatically when the virtual
	 * machine closes.
	 */
	private static void lock() {
		StringBuilder lockFileName = new StringBuilder(System.getProperty("user.home"));
		if (!System.getProperty("user.home").endsWith(File.separator))
			lockFileName.append(File.separator);
		lockFileName.append("Boa.lock");
		File lockFile = new File(lockFileName.toString());
		if (lockFile.exists()) {
			System.err.println("Another instance of this program is already running.\n\n" +
					"If you are absolutely certain that this is not the case you can remove " +
					lockFile.getAbsoluteFile() + " at your own risk.");
			System.exit(ANOTHER_INSTANCE_RUNNING_ERROR);
		} else {
			try {
				lockFile.createNewFile();
			} catch (IOException e) {
				System.err.println("Users home directory '" + 
						lockFile.getAbsoluteFile() +
						 "' not writable: " + e.getMessage());
				System.exit(TMP_DIRECTORY_NOT_WRITABLE);
			}
			lockFile.deleteOnExit();
		}
	}

	/**
	 * Prints usage information of this user interface
	 */
	private static void printUsage() {
		System.out.println(
			"\nUsage: Boa [OPTIONS]\n\n" +
			"Transforms data sources to output files as specified in the configuration\n" +
			"files.\n\n" +
			"  -h, --help                    Displays this help\n" +
			"  -v, --verbose                 Sets verbose output\n" +
			"  -m, --mainconfig FILENAME     Sets main configuration file to FILENAME\n");
	}
}
