/**
 * 
 */
package org.qriterium.util;

import com.megginson.sax.DataWriter;
import java.io.File;
import java.io.IOException;
import java.io.FileWriter;
import java.util.TreeMap;
import java.util.logging.Logger;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import org.qriterium.connection.QConnection;
import org.qriterium.connection.QConnectionFactory;
import org.qriterium.connection.UnsupportedConnectionTypeException;
import org.qriterium.io.Storable;
import org.qriterium.statement.QHeader;
import org.qriterium.statement.QStatements;
import org.qriterium.statement.RawConnection;
import org.xml.sax.SAXException;

/**
 * Receives RawConnections and creates QConnections when possible
 * @author Arto
 */
public class QConnectionHandler implements Storable {	
	public static final String XMLELEMENT_ROOT = "rawconnections";
	public static final String XMLELEMENT_CONNECTION = "connection";
	public static final String XMLELEMENT_RAWID = "rawid";
	//private static final String XMLELEMENT_DATESTAMP = "datestamp";
	public static final String XMLELEMENT_CREATEDAT = "created_at";
	public static final String XMLELEMENT_FROMRAWID = "fromrawid";
	public static final String XMLELEMENT_TORAWID = "torawid";
	public static final String XMLELEMENT_CONNECTIONTYPE = "connectiontype";
	public static final String XMLELEMENT_HASH = "hash";
	
	private static final String KEYUNIQUER = ".org.qriterium.";
	
	private QStatements statements = null;
	private TreeMap<String,RawConnection> rawConnections = null;
	
	private Logger logger = 
		Logger.getLogger("org.qriterium.util.QConnectionHandler");

	public QConnectionHandler(QStatements statements) {
		this.rawConnections = new TreeMap<String,RawConnection>();
		this.statements = statements;
	}
	
	/**
	 * Adds a new RawConnection.
	 * @param connection
	 */
	public void addRawConnection(RawConnection connection) {
		if (connection == null) {
			return;
		}
		this.rawConnections.put(nextUniqueKey(connection), connection);
	}
	
	/**
	 * Tries to create new QConnections.
	 * If given QStatement allows QConnectionHandler to create new QConnection
	 * objects, it does it and returns the number of QConnections created.
	 * If none were created, returns 0.
	 * @param rawId Statement's raw id
	 * @return Number of QConnections created
	 */
	public int statementAdded(String rawId) {
		if (this.statements == null) {
			return 0;
		}

		int createCount = 0;
		int hashIndex = 0;
		RawConnection rc;
		//System.out.println(this.rawConnections);
		while ((rc = this.rawConnections.remove(uniqueKey(rawId, 
		                                                                            hashIndex++))) != null) {

			Long fromId = this.statements.getStatementIdByRawId(rc.getFrom());
			Long toId = this.statements.getStatementIdByRawId(rc.getTo());
			if (fromId == null || toId == null) {
				// source-statement is not known, skipping.
				// this should not happen in real life.
				continue;
			}
			
			QConnection connection = QConnectionFactory.getInstance(rc.getType());
			if (connection == null) {
				// this is non-fatal, the program may continue.
                // this block should not be reached.
				new UnsupportedConnectionTypeException();
				continue;
			}

			connection.setHeader(new QHeader(IdFactory.getUniqueId(),
											                     rc.getRawId(),
											                     rc.getHash(),
											                     rc.getAt())); 
			connection.setFrom(fromId);
			connection.setTo(toId);
			//System.out.println("creating " + connection);
			//System.out.println(this.rawConnections);
			this.statements.addQStatement(connection);
			createCount++;
		}
		return createCount;
	}

	/**
	 * Stores the object's state in the given file.
	 * 
	 * @param location File where the state is to be saved
	 * @return true if saving was succesfully done, otherwise false
	 */
	public boolean store(String filename) {
		logger.info("Storing connection states to " + filename);
		
		File cacheFile = new File(filename);
		try {
			DataWriter writer = new DataWriter(new FileWriter(cacheFile, false));
			writer.setIndentStep(XMLProperties.INDENT);
			writer.startDocument();
			writer.startElement(XMLELEMENT_ROOT);
			for (RawConnection rc : this.rawConnections.values()) {
				writer.startElement(XMLELEMENT_CONNECTION);
				writer.dataElement(XMLELEMENT_RAWID, rc.getRawId());
				//writer.dataElement(XMLELEMENT_DATESTAMP, QHeader.DATEFORMAT.format(rc.getDatestamp()));
				writer.dataElement(XMLELEMENT_CREATEDAT, QHeader.DATEFORMAT.format(rc.getAt()));
				writer.dataElement(XMLELEMENT_FROMRAWID, rc.getFrom());
				writer.dataElement(XMLELEMENT_TORAWID, rc.getTo());
				writer.dataElement(XMLELEMENT_CONNECTIONTYPE, rc.getType());
				writer.dataElement(XMLELEMENT_HASH, rc.getHash());
				writer.endElement(XMLELEMENT_CONNECTION);
			}
			writer.endElement(XMLELEMENT_ROOT);
			writer.endDocument();
			return true;
		} catch (IOException e) {
			logger.severe("Could not store QConnectionHandler cache: " +
						e.getMessage());
			return false;
		} catch (org.xml.sax.SAXException e) {
			logger.severe("Could not store QConnectionHandler cache: " +
						e.getMessage());
			return false;
		}
	}
	
	/**
	 * Retrieves the object's state from the given file
	 * @param location The location where the state is to be retrieved
	 * @return true if retrieving was succesfully done, otherwise false
	 */	
	public boolean retrieve(String filename) {
		logger.info("Retrieving connection states from " + filename);
		
		File cacheFile = new File(filename);
		if (!cacheFile.exists()) {
			return false;
		}
		
		this.rawConnections = new TreeMap<String,RawConnection>();

		SAXParser saxParser;
		try {
			saxParser = SAXParserFactory.newInstance().newSAXParser();
			QConnectionHandlerCacheHandler myHandler 
				= new QConnectionHandlerCacheHandler(this, this.rawConnections);
			saxParser.parse(cacheFile, myHandler);
		} catch (ParserConfigurationException e) {
			logger.warning("Could not read QConnectionHandler cache: " +
						e.getMessage());
			return false;
		} catch (SAXException e) {
			logger.warning("Could not read QConnectionHandler cache: " +
						e.getMessage());
			return false;
		} catch (IOException e) {
			logger.warning("Could not read QConnectionHandler cache: " +
						e.getMessage());
			return false;
		}

		return true;
	}
	
	/**
	 * For unit testing
	 * @return Number of RawConnection objects stored by this instance.
	 */
	public int getHashSize() {
		return this.rawConnections.size();
	}
	
	/**
	 * For unit testing
	 * @return RawConnection objects stored by this instance
	 */
	public RawConnection[] getRawConnections() {
		if (this.getHashSize() == 0) {
			return new RawConnection[0];
		}
		RawConnection[] conns = new RawConnection[this.getHashSize()];
		return this.rawConnections.values().toArray(conns); 
	}
	
	/*
	 * A little helper to create hash keys. 
	 */
	private String uniqueKey(String baseKey, int count) {
		//System.out.println("got key: blaa."+count);
		return new String(baseKey + KEYUNIQUER + count);
	}
	
	private String nextUniqueKey(RawConnection connection) {
		return nextUniqueKey(connection.getTo());
	}
	
	/*
	 * A little helper to determine next suitable key for a RawConnection
	 */
	public String nextUniqueKey(String baseKey) {
		int count = 0;
		String key;
		while (this.rawConnections.containsKey(key = uniqueKey(baseKey, count))) {
			count++;
		}
		return key;
	}

}
