package org.qriterium.util;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.OutputStreamWriter;
import java.nio.charset.Charset;
import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;
import java.util.logging.Logger;

import com.megginson.sax.DataWriter;

import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.qriterium.io.*;
import org.qriterium.statement.QStatement;
import org.xml.sax.SAXException;


/**
 *
 * QStatementIndex indexes a String rawid to a QStatement id.
 * It is used for finding out if a QDocument is already known.
 * @author Ilmari
 * @author Arto
 */
public class QStatementIndex implements Storable
{
	public static final String XMLELEMENT_ROOT = "statementindex";
	public static final String XMLELEMENT_MAP = "map";
	public static final String XMLELEMENT_ID = "id";
	public static final String XMLELEMENT_RAWID = "rawid";

	public HashMap<String, Long> index;
	
	private Logger logger = Logger.getLogger("org.qriterium.util.QStatementIndex");
	
	/**
	 * Creates a new QStatementIndex.
	 */
	public QStatementIndex()
	{
		this.index = new HashMap<String, Long>();
	}
		
	/**
	 * Puts the QStatement's id into the index keyed by its rawId.
	 * 
	 * @param statement
	 */
	public void put(QStatement statement)
	{
		this.putRaw(statement.getRawId(), statement.getId());
	}

	/**
	 * Puts id into the index keyed by rawid.
	 * 
	 * @param rawid
	 * @param qid
	 */
	public void putRaw(String rawid, Long qid)
	{
		this.index.put(rawid, qid);
	}
	
	/**
	 * Get the QStatement id indexed by rawId.
	 * If no such QStatement id exists, returns null.
	 */
	public Long get(String rawId)
	{
		return this.index.get(rawId);
	}

	/**
	 * Writes the QStatementIndex to cacheFile.
	 * 
	 * @param cacheFile File to write index into.
	 * @return True on success, false on failure.
	 */
	public boolean store(String filename)
	{
		logger.info("Storing index state to " + filename);
		File cacheFile = new File(filename);
		
		FileOutputStream out;
		try {
			out = new FileOutputStream( cacheFile );
			BufferedWriter writer;
			writer = new BufferedWriter(
					new OutputStreamWriter(out, Charset.forName("UTF-8"))
			);
			DataWriter w = new DataWriter(writer);
			w.startDocument();
			w.startElement(XMLELEMENT_ROOT);
			
			Iterator< Map.Entry<String,Long> > it = this.index.entrySet().iterator();
			Map.Entry<String,Long> map;
			Long qid;
			String rawid;
			
			while(it.hasNext())
			{
				map = it.next();
				rawid = map.getKey();
				qid = map.getValue();
				writeXMLMap(w, rawid, qid);
			}
			
			w.endElement(XMLELEMENT_ROOT);
			w.endDocument();
			writer.close();
			out.close();
		} catch (Exception e) {
			logger.severe("Could not store QConnectionHandler cache: " +
					  e.getMessage());
			return false;
		}
		return true;
	}

	/**
	 * Reads the QStatementIndex from cacheFile.
	 * 
	 * @param cacheFile File to read index from.
	 * @return True on success, false on failure.
	 */
	public boolean retrieve(String filename)
	{
		logger.info("Retrieving index state from " + filename);
		File cacheFile = new File(filename); 
		 
		SAXParser saxParser;
		try {
			saxParser = SAXParserFactory.newInstance().newSAXParser();
			QStatementIndexCacheHandler myHandler 
				= new QStatementIndexCacheHandler(this);
			saxParser.parse(cacheFile, myHandler);
		} catch (Exception e) {
			logger.warning("Could not read QConnectionHandler cache: " +
						  e.getMessage());
			return false;
		}
		return true;
	}
	
	
	/**
	 * Dumps rawid and qid to disk as an XML element.
	 */
	private void writeXMLMap(DataWriter w, String rawid, Long qid) 
	throws SAXException
	{
		w.startElement(XMLELEMENT_MAP);
		w.dataElement(XMLELEMENT_ID, qid.toString());
		w.dataElement(XMLELEMENT_RAWID, rawid);
		w.endElement(XMLELEMENT_MAP);
	}
	
}

